/********************************************************
 *
 *  Copyright (c) 2015 Sony Corporation
 *
 ********************************************************/

#include <linux/mm.h>
#include <linux/poll.h>
#include <linux/module.h>
#include <linux/interrupt.h>

#include "exttype.h"
#define KDEBUG
#include "general_macro.h"

#include "msgdma_reg.h"
#include "msgdma_desc.h"
#include "fmd_dma.h"

MODULE_LICENSE("Dual BSD/GPL");

static MsgdmaReg_t __iomem*         msgdma_reg;
static MsgdmaDescriptor_t __iomem*  msgdma_desc;

static u32 dbg_counter_intr1 = 0;
static u32 dbg_counter_intr2 = 0;

DECLARE_WAIT_QUEUE_HEAD(dma_wait_queue);

void fmd_dma_initialize(u32 pci_addr)
{
    msgdma_reg  = (MsgdmaReg_t*)       (pci_addr + IO_MAP_OFFSET_CSR);
    msgdma_desc = (MsgdmaDescriptor_t*)(pci_addr + IO_MAP_OFFSET_DESCRIPTOR);
}

void fmd_dma_finalize(void)
{
    iowrite32(MSGDMA_SCR_STS_MASK, &msgdma_reg->status);
    iowrite32(MSGDMA_SCR_STS_MASK, &msgdma_reg->status);
}

int fmd_dma_reset(u32 bit)
{
    u32 status;
    // step 1.
    iowrite32(MSGDMA_SCR_CTRL_STOP, &msgdma_reg->control);

    // step 2.
    status = ioread32be(&msgdma_reg->fpga_status);

    if (status & (FPGA_STS_BUSY_DDR_W | FPGA_STS_BUSY_DDR_R)) {
        udelay(1000*1000);
        status = ioread32be(&msgdma_reg->fpga_status);
        if (status & (FPGA_STS_BUSY_DDR_W | FPGA_STS_BUSY_DDR_R)) {
            printk("%s %d: DDR Busy!!\n", __func__, __LINE__);
            //return ERROR;
        }
    }

    // step 3.
    iowrite32be(bit,  &msgdma_reg->fpga_reset);
    udelay(100*1000);
    iowrite32be(0x00000000,  &msgdma_reg->fpga_reset);

    // stap 4.
    iowrite32(MSGDMA_SCR_STS_IRQ, &msgdma_reg->status);
    fmd_msgdma_reset_and_wait_ready();

    return OK;
}

void fmd_dma_init_intr(void)
{
    iowrite32(0x00000000,               &msgdma_reg->status);
    iowrite32(MSGDMA_SCR_CTRL_INT_EN,   &msgdma_reg->control);
    iowrite32(0x00000000,               &msgdma_reg->status);
    iowrite32(MSGDMA_SCR_CTRL_INT_EN,   &msgdma_reg->control);
}

irqreturn_t fmd_dma_intr_msi(int irq, void* data)
{
    boolean is_fire = FALSE;

    u32 status = ioread32(&msgdma_reg->status);

    if (status & MSGDMA_SCR_STS_IRQ) {
        wake_up_interruptible(&dma_wait_queue);
        is_fire = TRUE;
        ++ dbg_counter_intr1;
    }

    if (is_fire) {
        return IRQ_HANDLED;
    }
    else {
        printk("%s: none interrupt. \n", __func__);
        return IRQ_NONE;
    }
}

void fmd_msgdma_finalize(void)
{
    iowrite32((MSGDMA_SCR_CTRL_STOP|MSGDMA_SCR_CTRL_STOP_DESC), &msgdma_reg->control);
    udelay(1*1000);
    iowrite32(MSGDMA_SCR_STS_MASK,   &msgdma_reg->status);
    iowrite32(MSGDMA_SCR_CTRL_RESET, &msgdma_reg->control);
    udelay(1*1000);
    iowrite32(0, &msgdma_reg->control);
}

int fmd_msgdma_reset_and_wait_ready(void)
{
    u32 status;
    u32 count = 0;

    iowrite32(MSGDMA_SCR_STS_MASK, &msgdma_reg->status);
    status = ioread32(&msgdma_reg->status);

    //kdebug("status    : 0x%08x\n", status);

    iowrite32(MSGDMA_SCR_CTRL_RESET, &msgdma_reg->control);

    do {
        status = ioread32(&msgdma_reg->status);
        udelay(1*1000);

        if (!(status & MSGDMA_SCR_STS_RESETTING) && !(status & MSGDMA_SCR_STS_BUSY)) {
            return OK;
        }
        count++;

    } while (count < 1000);

    return -EIO;
}


static const u32 DMA_BUF_SIZE  = 0x40000; //256K
static const u32 DMA_ALIGN     = 0x1000;  //4096
size_t fmd_dma_get_safty_size(size_t size)
{
    return (size % DMA_ALIGN == 0) ? size : (size + DMA_ALIGN) & 0xfffff000;
    //return size;
}

static const u32 DMA_RETRY_COUNT = 10;
static int fmd_msgdma_desc_ready(void)
{
    u32 status;
    u32 counter = 0;

    do {
        if (2 <= counter) {
            if (DMA_RETRY_COUNT <= counter) {
                return ERROR;
            }
            else {
                udelay(1*1000);
            }
        }

        status = ioread32(&msgdma_reg->status);
        counter ++;
    } while (status & MSGDMA_SCR_STS_DESC_FULL);

    return OK;
}

int fmd_dma_put_descriptor(u32 dma_paddr, u32 mask, u32 offset, size_t size, boolean is_read)
{
    u32     addr_fm;
    u32     addr_ram;
    u32     control;
    u32     num_of_buf = 0;
    size_t  total_size = fmd_dma_get_safty_size(size);
    size_t  write_size;

    //kdebug("total_size : %04x \n", total_size);
    //kdebug("read_fill : 0x%04x\n", ioread16(&msgdma_reg->read_fill ));
    //kdebug("write_fill: 0x%04x\n", ioread16(&msgdma_reg->write_fill));

    while (total_size != 0) {

        if (OK != fmd_msgdma_desc_ready()) {
            printk("%s: MSGDMA_SCR_STS_DESC_FULL\n", __func__);
            return ERROR;
        }

        addr_fm  = IO_MAP_OFFSET_FPGA_DDR + offset + (num_of_buf * DMA_BUF_SIZE);
        addr_ram = (dma_paddr + (num_of_buf * DMA_BUF_SIZE)) & ~mask;

        if (is_read) {
            iowrite32(addr_fm,  &msgdma_desc->read_addr);
            iowrite32(addr_ram, &msgdma_desc->write_addr);
        }
        else {
            iowrite32(addr_ram, &msgdma_desc->read_addr);
            iowrite32(addr_fm,  &msgdma_desc->write_addr);
        }

        if (DMA_BUF_SIZE < total_size) {
            write_size = DMA_BUF_SIZE;
            total_size -= DMA_BUF_SIZE;
        }
        else {
            write_size = total_size;
            total_size = 0;
        }
        iowrite32(write_size, &msgdma_desc->length);

        control = MSGDMA_DESC_CTL_TR_ERR_IRQ | MSGDMA_DESC_CTL_GO;
        if (total_size == 0) {
            control |= MSGDMA_DESC_CTL_TR_COMP_IRQ;
            ++ dbg_counter_intr2;
        }
        iowrite32(control, &msgdma_desc->control);

        num_of_buf++;
    }

    return OK;
}

int fmd_dma_wait(void)
{
    int ret_val = wait_event_interruptible_timeout(
                    dma_wait_queue,
                    (MSGDMA_SCR_STS_IRQ & ioread32(&msgdma_reg->status)),
                    2*1000);

    // Clear INT bit
    iowrite32(MSGDMA_SCR_STS_IRQ, &msgdma_reg->status);

    if (ret_val == 0) {
        printk("%s: timeout. \n", __func__);
        fmd_dma_show_status();
        return -EBUSY;
    }
    else if (ret_val < 0) {
        printk("%s: fault. \n", __func__);
        return -EFAULT;
    }

    return OK;
}

int fmd_dma_read(char* dma_vbuf, char __user* buf, size_t size)
{
    u32     num_of_buf = 0;
    size_t  total_size = size;

    while (DMA_BUF_SIZE <= total_size) {
        if (copy_to_user(buf + (num_of_buf * DMA_BUF_SIZE),
                         dma_vbuf + (num_of_buf * DMA_BUF_SIZE),
                         DMA_BUF_SIZE)) {
            return -EFAULT;
        }

        num_of_buf++;
        total_size -= DMA_BUF_SIZE;
    }

    if (total_size != 0) {
        if (copy_to_user(buf + (num_of_buf * DMA_BUF_SIZE),
                         dma_vbuf + (num_of_buf * DMA_BUF_SIZE),
                         total_size)) {
            return -EFAULT;
        }
        num_of_buf++;
    }
    return OK;
}

ssize_t fmd_dma_write(char* dma_vbuf, const char __user* buf, size_t size)
{
    u32     num_of_buf = 0;
    size_t  total_size = size;

    while (DMA_BUF_SIZE <= total_size) {
        if (copy_from_user(dma_vbuf + (num_of_buf * DMA_BUF_SIZE),
                           buf + (num_of_buf * DMA_BUF_SIZE),
                           DMA_BUF_SIZE)) {
            return -EFAULT;
        }

        num_of_buf++;
        total_size -= DMA_BUF_SIZE ;
    }

    if (total_size != 0) {
        if (copy_from_user(dma_vbuf + (num_of_buf * DMA_BUF_SIZE),
                           buf + (num_of_buf * DMA_BUF_SIZE),
                           total_size)) {
            return -EFAULT;
        }
        num_of_buf++;
    }

    return OK;
}

int fmd_dma_get_status(void)
{
    u32 dma_status  = ioread32(&msgdma_reg->status);
    u32 fpga_status = ioread32be(&msgdma_reg->fpga_status);

    return ((dma_status << 8) | fpga_status);
}

int fmd_dma_show_status(void)
{
    u32 status = ioread32(&msgdma_reg->status);

    printk("[ mSGDMA CSR ]\n");
    printk("status    : 0x%08x\n", status);
    printk("    BUSY       : %s\n", status & MSGDMA_SCR_STS_BUSY       ? "O" : " ");
    printk("    DESC_EMPTY : %s\n", status & MSGDMA_SCR_STS_DESC_EMPTY ? "O" : " ");
    printk("    DESC_FULL  : %s\n", status & MSGDMA_SCR_STS_DESC_FULL  ? "O" : " ");
    printk("    RESP_EMPTY : %s\n", status & MSGDMA_SCR_STS_RESP_EMPTY ? "O" : " ");
    printk("    RESP_FULL  : %s\n", status & MSGDMA_SCR_STS_RESP_FULL  ? "O" : " ");
    printk("    STOPPED    : %s\n", status & MSGDMA_SCR_STS_STOPPED    ? "O" : " ");
    printk("    RESETTING  : %s\n", status & MSGDMA_SCR_STS_RESETTING  ? "O" : " ");
    printk("    ERROR      : %s\n", status & MSGDMA_SCR_STS_ERROR      ? "O" : " ");
    printk("    EL_TERM    : %s\n", status & MSGDMA_SCR_STS_EL_TERM    ? "O" : " ");
    printk("    IRQ        : %s\n", status & MSGDMA_SCR_STS_IRQ        ? "O" : " ");
    printk("control   : 0x%08x\n", ioread32(&msgdma_reg->control   ));
    printk("read_fill : 0x%04x\n", ioread16(&msgdma_reg->read_fill ));
    printk("write_fill: 0x%04x\n", ioread16(&msgdma_reg->write_fill));
    printk("read seq  : 0x%04x\n", ioread16(&msgdma_reg->r_seq_num));
    printk("write seq : 0x%04x\n", ioread16(&msgdma_reg->w_seq_num));

    printk("counter   : 0x%08x\n", dbg_counter_intr1);
    printk("counter   : 0x%08x\n", dbg_counter_intr2);
    printk("\n");

    status = ioread32be(&msgdma_reg->fpga_status);

    printk("[ FPGA Status ]\n");
    printk("fpgastatus: 0x%08x\n", status);
    printk("    BUSY_DDR_W : %s\n", status & FPGA_STS_BUSY_DDR_W ? "O" : " ");
    printk("    BUSY_DDR_R : %s\n", status & FPGA_STS_BUSY_DDR_R ? "O" : " ");
    printk("    BUSY_QUEUE : %s\n", status & FPGA_STS_BUSY_QUEUE ? "O" : " ");
    printk("    BUSY_CACHE : %s\n", status & FPGA_STS_BUSY_CACHE ? "O" : " ");
    printk("\n");

    return OK;
}
