/********************************************************
 *
 *  Copyright (c) 2017 Sony Corporation
 *
 ********************************************************/

#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/cdev.h>
#include <linux/mm.h>
#include <linux/poll.h>

#include "exttype.h"
#define KDEBUG
#include "general_macro.h"
#include "dbg_assert.h"

#include "fmdriver.h"
#include "msgdma_reg.h"
#include "msgdma_desc.h"
#include "fmd_dma.h"

#define DEVICE_ID    0xe001
#define VENDOR_ID    0x1172

MODULE_LICENSE("Dual BSD/GPL");

static int fm_devs  = 1;
static int fm_major = 0;
module_param(fm_major, uint, 0);

static struct pci_device_id fmd_pci_tbl[] = {
    {VENDOR_ID, DEVICE_ID, PCI_ANY_ID, PCI_ANY_ID, 0, 0, },
    {0,}
};

MODULE_DEVICE_TABLE(pci, fmd_pci_tbl);

typedef enum bar {
     //e_bar0 = 0,
     //e_bar1,
     e_bar2 = 0,
     e_bar3,
     e_bar_max
} bar_t;

typedef struct {
    uchar*  addr;
    ulong   size;
} area_info_t;

typedef struct pci_dev_data {
    struct pci_dev* pdev;
    area_info_t     area[2];
} pci_info_t;

typedef struct {
    u32        paddr;
    void*      vaddr;
    u32        size;
} dma_set_t;

static pci_info_t               pcinfo;
static struct cdev              fm_cdev;

static u32                          a2p_mask;
static dma_set_t                    dma_set;
static PcieDdrReg_t __iomem*        ddr_reg;
static PcieCraReg_t __iomem*        cra_reg;

static int            fmd_pci_probe(struct pci_dev *pdev, const struct pci_device_id* ent);
static void __devexit fmd_pci_remove(struct pci_dev* pdev);

static struct pci_driver fm_driver = {
    .name     = DEVICE_NAME,
    .id_table = fmd_pci_tbl,
    .probe    = fmd_pci_probe,
    .remove   = __devexit_p(fmd_pci_remove),
};

static void fmd_pci_intx_disable(void)
{
#define PCI_COMMAND_DISABLE_INTx 0x400
    u16 vendor_id;
    u16 device_id;
    u16 command;

    pci_read_config_word(pcinfo.pdev, PCI_VENDOR_ID, &vendor_id);
    pci_read_config_word(pcinfo.pdev, PCI_DEVICE_ID, &device_id);
    pci_read_config_word(pcinfo.pdev, PCI_COMMAND,   &command);

    command |= (PCI_COMMAND_DISABLE_INTx);
    pci_write_config_word(pcinfo.pdev, PCI_COMMAND, command);

    pci_set_master(pcinfo.pdev);

    pci_read_config_word(pcinfo.pdev, PCI_COMMAND, &command);

    kdebug("VENDOR_ID : %04x \n", vendor_id);
    kdebug("DEVICE_ID : %04x \n", device_id);
    kdebug("COMMAND   : %04x \n", command);
}

static void fmd_pci_show_config(void)
{
    u16 vendor_id;
    u16 device_id;
    u16 command;

    pci_read_config_word(pcinfo.pdev, PCI_VENDOR_ID, &vendor_id);
    pci_read_config_word(pcinfo.pdev, PCI_DEVICE_ID, &device_id);
    pci_read_config_word(pcinfo.pdev, PCI_COMMAND,   &command);

    printk("=== [ PCI Config ] ===\n");
    printk("VENDOR_ID : %04x \n", vendor_id);
    printk("DEVICE_ID : %04x \n", device_id);
    printk("COMMAND   : %04x \n", command);
    printk("\n");
}

static void fmd_pci_cra_addr_trans_tbl_initialize(void)
{
    u32 a2p;

    iowrite32(0xfffffffc, &cra_reg->a2p_addr_l0);
    a2p_mask = ioread32(&cra_reg->a2p_addr_l0);
    iowrite32(dma_set.paddr & a2p_mask, &cra_reg->a2p_addr_l0);
    a2p = ioread32(&cra_reg->a2p_addr_l0);

    kdebug("mask        : 0x%08x\n", a2p_mask);
    kdebug("a2p(before) : 0x%08x\n", a2p);
    iowrite32(0x00000000, &cra_reg->a2p_addr_h0);

    iowrite32(a2p + 0x01000000, &cra_reg->a2p_addr_l1);
    iowrite32(0x00000000, &cra_reg->a2p_addr_h1);
    a2p = ioread32(&cra_reg->a2p_addr_l1);
    kdebug("a2p(after)  : 0x%08x\n", a2p);
}

static void fmd_pci_cra_addr_trans_tbl_finalize(void)
{
    iowrite32(0, &cra_reg->a2p_addr_l0);
    iowrite32(0, &cra_reg->a2p_addr_h0);
    iowrite32(0, &cra_reg->a2p_addr_l0);
    iowrite32(0, &cra_reg->a2p_addr_h0);
}

static void fmd_pci_cra_irq_initialize(void)
{
    iowrite32(0x0000ffff, &cra_reg->cra_irq_control);
}

static void fmd_pci_cra_dma_reset(void)
{
    iowrite32(ULONG_MAX,  &cra_reg->dma_reset);
    udelay(1*1000);
    iowrite32(0x00000000, &cra_reg->dma_reset);
}

static void fmd_pci_cra_show(void)
{
    printk("=== [ CRA ] ===\n");
    printk("IRQ STATUS : %08x\n", ioread32(&cra_reg->cra_irq_status));
    printk("IRQ_CONTROL: %08x\n", ioread32(&cra_reg->cra_irq_control));
    printk("A2P_ADDR_L0: %08x\n", ioread32(&cra_reg->a2p_addr_l0));
    printk("A2P_ADDR_H0: %08x\n", ioread32(&cra_reg->a2p_addr_h0));
    printk("A2P_ADDR_L1: %08x\n", ioread32(&cra_reg->a2p_addr_l1));
    printk("A2P_ADDR_H1: %08x\n", ioread32(&cra_reg->a2p_addr_h1));
    printk("\n");
}

static const u32 BIGSIZE_ORDER = 10;
static int fmd_dma_buf_initialize(void)
{
    dma_set.vaddr = (void*)__get_free_pages(GFP_KERNEL | GFP_DMA, BIGSIZE_ORDER);
    dma_set.size  = PAGE_SIZE*((2<<BIGSIZE_ORDER)/2);
    dma_set.paddr = __pa(dma_set.vaddr);

    if (dma_set.vaddr == NULL) {
        printk("%s: unable to alloc dma mem.\n", __func__);
        return -ENOMEM;
    }

    kdebug("vaddr : 0x%08x\n", (u32)dma_set.vaddr);
    kdebug("paddr : 0x%08x\n", (u32)__pa(dma_set.vaddr));
    kdebug("paddr : 0x%08llx\n", (u64)dma_set.paddr);
    kdebug("shift : 0x%08x\n", (u32)(dma_set.paddr) >> PAGE_SHIFT);

    printk("dma size  : 0x%08x\n", dma_set.size);

    return OK;
}

static int fmd_dma_buf_show(void)
{
    printk("=== [ DMA ] ===\n");
    printk("vaddr : 0x%08x\n", (u32)dma_set.vaddr);
    printk("paddr : 0x%08x\n", (u32)__pa(dma_set.vaddr));
    printk("shift : 0x%08x\n", (u32)(dma_set.paddr) >> PAGE_SHIFT);
    printk("size  : 0x%08x\n", dma_set.size);
    printk("\n");

    return OK;
}

//------------------------------------------------------------------
static int fmd_pci_setup(void)
{
    int ret;

    DBG_ASSERT_CONST_CONDITION(sizeof(MsgdmaReg_t)        == 0x80);
    DBG_ASSERT_CONST_CONDITION(sizeof(MsgdmaDescriptor_t) == 0x10);
    DBG_ASSERT_CONST_CONDITION(sizeof(dma_addr_t)         == 0x08);
    DBG_ASSERT_CONST_CONDITION(sizeof(PcieCraReg_t)       == 0x5030);
    DBG_ASSERT_CONST_CONDITION(sizeof(ioctl_set_t)        == 16);
    DBG_ASSERT_CONST_CONDITION(OFFSET_OF(PcieCraReg_t, cra_irq_status)  == 0x0040);
    DBG_ASSERT_CONST_CONDITION(OFFSET_OF(PcieCraReg_t, cra_irq_control) == 0x0050);
    DBG_ASSERT_CONST_CONDITION(OFFSET_OF(PcieCraReg_t, a2p_addr_l0)     == 0x1000);
    DBG_ASSERT_CONST_CONDITION(OFFSET_OF(PcieCraReg_t, dma_reset)       == 0x4050);
    DBG_ASSERT_CONST_CONDITION(OFFSET_OF(PcieCraReg_t, intr_data)       == 0x4F00);
    DBG_ASSERT_CONST_CONDITION(OFFSET_OF(PcieCraReg_t, intr_enable)     == 0x5000);
    DBG_ASSERT_CONST_CONDITION(OFFSET_OF(PcieCraReg_t, rx_ctrl)         == 0x5010);
    DBG_ASSERT_CONST_CONDITION(OFFSET_OF(PcieCraReg_t, tx_ctrl)         == 0x5020);
    DBG_ASSERT_CONST_CONDITION(OFFSET_OF(PcieDdrReg_t, ctrl)         == 0x0000);
    DBG_ASSERT_CONST_CONDITION(OFFSET_OF(PcieDdrReg_t, size)         == 0x0020);
    DBG_ASSERT_CONST_CONDITION(OFFSET_OF(PcieDdrReg_t, addr_u)       == 0x0030);
    DBG_ASSERT_CONST_CONDITION(OFFSET_OF(PcieDdrReg_t, addr_l)       == 0x0040);

    cra_reg     = (PcieCraReg_t*)((u32)pcinfo.area[e_bar2].addr + IO_MAP_OFFSET_CRA);
    ddr_reg     = (PcieDdrReg_t*)((u32)pcinfo.area[e_bar2].addr + IO_MAP_OFFSET_DDR_CTRL);

    fmd_dma_initialize((u32)pcinfo.area[e_bar2].addr);

    ret = fmd_dma_buf_initialize();
    if (ret) {
        return ret;
    }

    fmd_pci_intx_disable();
    fmd_pci_cra_addr_trans_tbl_initialize();
    fmd_pci_cra_irq_initialize();

    ret = fmd_msgdma_reset_and_wait_ready();
    if (ret) {
        printk("%s: Failed to initialize mSGDMA\n", __func__);
        return ret;
    }

    ret = pci_enable_msi(pcinfo.pdev);
    if (ret) {
        printk("%s: Failed to set MSI Enable\n", __func__);
        return ret;
    }

    if (pcinfo.pdev->msi_enabled) {
        kdebug("MSI Enable\n");
    }

    ret = request_irq(pcinfo.pdev->irq,
                      &fmd_dma_intr_msi,
                      IRQF_SHARED|IRQF_TRIGGER_MASK,
                      DEVICE_NAME,
                      &pcinfo);

    if (ret) {
        printk("%s: Failed to set interrupt handler\n", __func__);
        return ret;
    }

    return ret;
}

static int __devinit fmd_pci_probe(struct pci_dev*               pdev,
                                   const struct pci_device_id*   id)
{
    if (pci_enable_device(pdev) != OK) {
        printk("%s: cannot enable the PCI device\n", __func__);
        return ERROR;
    }

    if (pci_resource_flags(pdev, 2) & IORESOURCE_MEM) {

        if (IOMAP_LENGTH > pci_resource_len(pdev, 2)) {
            printk("%s: BAR2 too large mapping size\n", __func__);
            return ERROR;
        }

        pcinfo.area[e_bar2].size = IOMAP_LENGTH;
        pcinfo.area[e_bar2].addr = pci_iomap(pdev, 2, 0);

        if (pcinfo.area[e_bar2].addr == NULL) {
            printk("%s: BAR0 cannot allocate base address\n", __func__);
            return ERROR;
        }
    }
    else {
        printk("%s: MMIO is not available\n", __func__);
        return ERROR;
    }

    pcinfo.pdev = pdev;

    printk("%s %s \n", __DATE__, __TIME__);

    fmd_pci_setup();

    return OK;
};

static void __devexit fmd_pci_remove(struct pci_dev* pdev)
{
    fmd_msgdma_finalize();

    iowrite32(0, &cra_reg->cra_irq_control);

    fmd_pci_cra_dma_reset();

    fmd_pci_cra_addr_trans_tbl_finalize();

    free_pages((int)dma_set.vaddr, BIGSIZE_ORDER);

    free_irq(pcinfo.pdev->irq, &pcinfo);

    pci_disable_msi(pcinfo.pdev);

    pci_iounmap(pdev, pcinfo.area[e_bar2].addr);

    pci_disable_device(pdev);
}

static int fmd_fops_open(struct inode* inode, struct file* file)
{
   //kdebug("%s: major %d minor %d (pid %d)\n",
   //       __func__, imajor(inode), iminor(inode), current->pid);

    inode->i_private = inode;
    file->private_data = file;

    return OK;
}

static int fmd_fops_close(struct inode* inode, struct file* file)
{
   //kdebug("%s: major %d minor %d (pid %d)\n",
   //       __func__, imajor(inode), iminor(inode), current->pid);

    return OK;
}

static loff_t fmd_fops_llseek(struct file * filep, loff_t offset, int whence)
{
    if (whence == SEEK_SET) {
        return offset;
    }
    else {
        return 0;
    }
}

static ssize_t fmd_fops_read(struct file* filep, char __user* buf, size_t count, loff_t* f_pos)
{
    int ret;

    (void)filep;

    fmd_dma_init_intr();

    ret = fmd_dma_put_descriptor(dma_set.paddr, a2p_mask, *f_pos, count, TRUE);
    if (ret) {
        return ret;
    }

    ret = fmd_dma_wait();
    if (ret) {
        return ret;
    }

    dma_unmap_single(&pcinfo.pdev->dev, dma_set.paddr, count, DMA_FROM_DEVICE);

    ret = fmd_dma_read(dma_set.vaddr, buf, count);
    if (ret) {
        return ret;
    }

    fmd_dma_finalize();

    return OK;
}

static ssize_t fmd_fops_write(struct file*          filep,
                              const char __user*    buf,
                              size_t                count,
                              loff_t*               f_pos)
{
    int ret;

    (void)filep;

    fmd_dma_init_intr();

    ret = fmd_dma_write(dma_set.vaddr, buf, count);
    if (ret) {
        return ret;
    }
    dma_map_single(&pcinfo.pdev->dev, dma_set.vaddr, count, DMA_TO_DEVICE);

    ret = fmd_dma_put_descriptor(dma_set.paddr, a2p_mask, *f_pos, count, FALSE);
    if (ret) {
        return ret;
    }

    ret = fmd_dma_wait();
    if (ret) {
        return ret;
    }

    fmd_dma_finalize();

    return OK;
}

static int fmd_fops_mmap(struct file* filep, struct vm_area_struct* vma)
{
    int ret = 0;

    if ((vma->vm_end - vma->vm_start) > dma_set.size) {
        printk("%s: vma size (0x%lx) is too big.\n", __func__ ,(vma->vm_end - vma->vm_start));
        return -EINVAL;
    }

    ret = remap_pfn_range(
            vma,                                // 仮想メモリ(VMA)構造体
            vma->vm_start,                      // 仮想メモリの開始番地
            (u32)(dma_set.paddr) >> PAGE_SHIFT, // 物理メモリのページ番号
            vma->vm_end - vma->vm_start,        // メモリサイズ
            vma->vm_page_prot);                 // ページアトリビューション

    if (ret) {
        printk("%s: remap_page_range() failed. %d\n",__func__, ret);
        return ret;
    }

    return OK;
}

static int fmd_ioctol_dma_ram_to_fm(u32 offset, size_t size)
{
    int ret = 0;

    fmd_dma_init_intr();

    dma_map_single(&pcinfo.pdev->dev, dma_set.vaddr, size, DMA_TO_DEVICE);

    ret = fmd_dma_put_descriptor(dma_set.paddr, a2p_mask, offset, size, FALSE);
    if (ret) {
        return ret;
    }

    ret = fmd_dma_wait();
    if (ret) {
        return ret;
    }

    fmd_dma_finalize();

    return OK;
}

static int fmd_ioctol_dma_fm_to_ram(u32 offset, size_t size)
{
    int ret;

    fmd_dma_init_intr();

    ret = fmd_dma_put_descriptor(dma_set.paddr, a2p_mask, offset, size, TRUE);
    if (ret) {
        return ret;
    }

    ret = fmd_dma_wait();
    if (ret) {
        return ret;
    }

    dma_unmap_single(&pcinfo.pdev->dev, dma_set.paddr, size, DMA_FROM_DEVICE);

    fmd_dma_finalize();

    return OK;
}

static int fmd_ioctol_set_ddr_param(u64 offset, size_t size)
{
    size_t total_size = fmd_dma_get_safty_size(size);

    u32 offset_u = ULLONG_H_ULONG(offset);
    u32 offset_l = ULLONG_L_ULONG(offset);

    iowrite32be(offset_u,   &ddr_reg->addr_u);
    iowrite32be(offset_l,   &ddr_reg->addr_l);
    iowrite32be(total_size, &ddr_reg->size);

    return OK;
}

static int fmd_ioctol_set_ddr_ctrl(u32 offset)
{
    iowrite32be(offset, &ddr_reg->ctrl);

    return OK;
}

static int fmd_ioctol_reg_write(u32 offset, size_t size, u8* data)
{
    u32 i;
    u8* ptr = (u8*)((u32)pcinfo.area[e_bar2].addr + IO_MAP_OFFSET_CRA);
    for (i = 0; i < size; ++i) {
        ptr[offset + i] = data[i];
        //printk("%02x %02x\n", data[i], ptr[offset + i]);
    }
    return OK;
}

static int fmd_ioctol_reg_read(u32 offset, size_t size, u8* data)
{
    u32 i;
    u8* ptr = (u8*)((u32)pcinfo.area[e_bar2].addr + IO_MAP_OFFSET_CRA);
    for (i = 0; i < size; ++i) {
        data[i] = ptr[offset + i];
    }

    return OK;
}

static int fmd_ioctol_show(void)
{
    fmd_dma_show_status();

    return OK;
}

static int fmd_ioctol_reset(u32 bit)
{
    // step 1.
    iowrite32be(0, &ddr_reg->ctrl);

    return fmd_dma_reset(bit);
}

static int fmd_ioctol_get_status(void)
{
    return fmd_dma_get_status();
}

int fmd_fops_ioctl(struct inode* inode, struct file* filep, uint cmd, ulong arg)
{
    int ret = 0;
    ioctl_set_t* ioctl_set;

    ioctl_set = (ioctl_set_t __user *)arg;

    if (!access_ok(VERIFY_READ, (struct ioctl_set __user *)arg, _IOC_SIZE(cmd))) {
        kdebug("%s %s %d\n", __FILE__, __FUNCTION__, __LINE__);
    }

    switch (cmd) {
    case IOCTL_READ:
        ret = fmd_ioctol_dma_fm_to_ram(ioctl_set->offset, ioctl_set->size);
        break;

    case IOCTL_WRITE:
        ret = fmd_ioctol_dma_ram_to_fm(ioctl_set->offset, ioctl_set->size);
        break;

    case IOCTL_INFO:
        ret = fmd_ioctol_show();
        break;

    case IOCTL_DDR_PARAM:
        ret = fmd_ioctol_set_ddr_param(ioctl_set->offset, ioctl_set->size);
        break;

    case IOCTL_DDR_CTRL:
        ret = fmd_ioctol_set_ddr_ctrl(ioctl_set->offset);
        break;

    case IOCTL_RESET:
        ret = fmd_ioctol_reset(ioctl_set->offset);
        break;

    case IOCTL_STATUS:
        ret = fmd_ioctol_get_status();
        break;

    case IOCTL_REG_WRITE:
        ret = fmd_ioctol_reg_write(ioctl_set->offset, ioctl_set->size, ioctl_set->data);
        break;

    case IOCTL_REG_READ:
        ret = fmd_ioctol_reg_read(ioctl_set->offset, ioctl_set->size, ioctl_set->data);
        break;

    default:
        printk("%s: Command should be read or write.\n", __func__);
        return ERROR;
    }

    return ret;
}

//------------------------------------------------------------------
static struct kobject* kobj_fmd;

static ssize_t fmd_sysfs_status_show(struct kobject *kobj, struct kobj_attribute *attr, char *buf)
{
    fmd_pci_show_config();

    fmd_dma_buf_show();

    fmd_pci_cra_show();

    fmd_ioctol_show();

    return 0;
}
#if 0
static ssize_t fmd_sysfs_reset_show(struct kobject *kobj, struct kobj_attribute *attr, char *buf)
{
    return sprintf(buf, "%s:%d\n", attr->attr.name, foo);
}

static ssize_t fmd_sysfs_reset_store(struct kobject *kobj, struct kobj_attribute *attr, const char *buf, size_t count)
{
    sscanf(buf, "%du", &foo);
    return count;
}
#endif

static struct kobj_attribute fmd_attr_status = __ATTR(fmd_status, 0444, fmd_sysfs_status_show, NULL);

static struct attribute *attrs[] = {
    &fmd_attr_status.attr,
    NULL, /* need to NULL terminate the list of attributes */
};

static struct attribute_group attr_group = {
    .attrs = attrs,
};

static int fmd_sysfs_create(void)
{
    int retval;

    /* /sys/kernel/fmd/fmd_status */
    kobj_fmd = kobject_create_and_add("fmd", kernel_kobj);
    if (!kobj_fmd)
        return -ENOMEM;

    retval = sysfs_create_group(kobj_fmd, &attr_group);
    if (retval)
        kobject_put(kobj_fmd);

    return retval;
}

static void fmd_sysfs_remove(void)
{
    kobject_put(kobj_fmd);
}

//------------------------------------------------------------------
struct file_operations fmd_fops = {
    .owner   = THIS_MODULE,
    .open    = fmd_fops_open,
    .release = fmd_fops_close,
    .read    = fmd_fops_read,
    .write   = fmd_fops_write,
    .mmap    = fmd_fops_mmap,
    .ioctl   = fmd_fops_ioctl,
    .llseek  = fmd_fops_llseek,
};

static int fmd_init(void)
{
    dev_t dev = MKDEV(fm_major, 0);
    int ret;

    ret = alloc_chrdev_region(&dev, 0, fm_devs, DEVICE_NAME);
    if (ret) {
        goto error_cdev_alloc;
    }
    fm_major = MAJOR(dev);

    cdev_init(&fm_cdev, &fmd_fops);
    fm_cdev.owner = THIS_MODULE;

    ret = pci_register_driver(&fm_driver);
    if (ret) {
        goto error_pci;
    }

    ret = cdev_add(&fm_cdev, MKDEV(fm_major, 0), fm_devs);
    if (ret) {
        goto error_cdev_add;
    }

    fmd_sysfs_create();

    printk("%s driver(major %d) installed.\n", DEVICE_NAME, fm_major);

    return 0;

error_cdev_add:
    pci_unregister_driver(&fm_driver);
error_pci:
    unregister_chrdev_region(dev, fm_devs);
error_cdev_alloc:

    return ret;
}

static void fmd_cleanup(void)
{
    dev_t dev = MKDEV(fm_major, 0);

    fmd_sysfs_remove();

    pci_unregister_driver(&fm_driver);

    cdev_del(&fm_cdev);

    unregister_chrdev_region(dev, fm_devs);

    printk("driver %s unloaded\n", DEVICE_NAME);
}

module_init(fmd_init);
module_exit(fmd_cleanup);
